"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AsyncReadStream = void 0;
const AsyncBlockingQueue_1 = require("./AsyncBlockingQueue");
class AsyncReadStream {
    constructor(socket) {
        this.chunks = new AsyncBlockingQueue_1.AsyncBlockingQueue();
        this.workingChunk = null;
        this.consumed = 0;
        // socket event 'data': assume the type of chunk is Buffer
        // so dont set encoding of the stream
        socket.on('data', (chunk) => {
            console.log(`[AsyncReadStream] Received: '${chunk.toString()}'`);
            this.chunks.enqueue(chunk); // needn't await
        });
        // socket event 'error': needn't interrupt all waiting operations
        socket.on('error', err => {
            console.error('[AsyncReadStream] Socket Error:', err);
        });
        // socket event 'close': interrupt all waiting operations
        socket.on('close', _e => {
            this.chunks.releaseBlockedEnqueue();
            this.chunks.releaseBlockedDequeue(new Error("read the closed socket"));
        });
    }
    // must continue a loop after invoking this method
    async nextChunk() {
        const chunk = await this.chunks.dequeue();
        this.workingChunk = chunk;
        this.consumed = 0;
    }
    noConsumed() {
        if (this.workingChunk === null) {
            return 0;
        }
        return this.workingChunk.length - this.consumed;
    }
    consume(delta) {
        if (this.workingChunk === null) {
            return "";
        }
        const noConsumed = this.noConsumed();
        if (noConsumed <= 0) {
            return "";
        }
        const finalDelta = delta > noConsumed ? noConsumed : delta;
        const ret = this.workingChunk.toString("utf-8", this.consumed, this.consumed + finalDelta);
        this.consumed += finalDelta;
        return ret;
    }
    async readBytesAsString(n) {
        let ret = "";
        while (n > 0) {
            const s = this.consume(n);
            if (s.length <= 0) {
                await this.nextChunk();
                continue;
            }
            ret += s;
            n -= s.length;
        }
        return ret;
    }
    async readByte() {
        while (true) {
            if (this.noConsumed() <= 0) {
                await this.nextChunk();
                continue;
            }
            if (this.workingChunk !== null) { // always true
                return this.workingChunk[this.consumed++];
            }
        }
    }
    async readLine() {
        const MAX_LINE_SIZE = 1024 * 12;
        const buf = Buffer.alloc(MAX_LINE_SIZE);
        let sz = 0;
        while (true) {
            const b = await this.readByte();
            if (b === 10) { // '\n'
                break;
            }
            if (b === 13) { // '\r'
                continue;
            }
            buf[sz++] = b;
            if (sz >= buf.length) {
                break;
            }
        }
        return buf.toString('utf-8', 0, sz);
    }
}
exports.AsyncReadStream = AsyncReadStream;
//# sourceMappingURL=AsyncStream.js.map