"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MobdebugSession = void 0;
const debugadapter_1 = require("@vscode/debugadapter");
const debugadapter_2 = require("@vscode/debugadapter");
const debugadapter_3 = require("@vscode/debugadapter");
const net = require("net");
const path_1 = require("path");
const AsyncBlockingQueue_1 = require("./AsyncBlockingQueue");
const AsyncStream_1 = require("./AsyncStream");
class MobdebugSession extends debugadapter_1.LoggingDebugSession {
    constructor() {
        super();
        this.server = null;
        this.client = null;
        this.clientReader = null;
        this.cmdQueue = new AsyncBlockingQueue_1.AsyncBlockingQueue();
        this.breakpointMap = {};
        this.baseDir = "";
        this.breakPointCount = 1;
    }
    initializeRequest(response, args) {
        console.log("[initializeRequest] called");
        response.body = {};
        this.sendResponse(response);
    }
    bindClient(client) {
        client.on('close', _hadErr => this.onDisconnected());
        client.on('error', e => this.sendErrorOutput(`lua client error ${e}`));
        this.server?.close(); // dont accept incoming connection
        this.server = null;
        this.client = client;
        this.clientReader = new AsyncStream_1.AsyncReadStream(client);
        this.startIOLoop();
    }
    // just unbind the client, this method still works when client is null
    unbindClient() {
        this.server?.close();
        this.server = null;
        this.client = null;
        this.clientReader = null;
        this.stopIOLoop();
    }
    // triggered when the remote socket connected
    onConnected(client) {
        console.log("[onConnected] called");
        this.sendOutput(`lua debugger client connected`);
        this.bindClient(client);
        this.runInitCommands();
    }
    // triggered when the remote socket closed
    onDisconnected() {
        console.log("[onDisconnected] called");
        this.sendOutput(`lua debugger client disconnected`);
        this.runLastCommands();
        this.unbindClient();
    }
    // triggered when the debugger session detached from vscode
    // note: vscode may detach the session before connecting to lua
    onDetach() {
        console.log("[onDetach] called");
        this.sendOutput(`debug session detached`);
        // client may be null when calling this method
        // note: unbindClient will set this.client to null, use cli instead
        const cli = this.client;
        cli?.end("DONE\n", () => cli?.destroy());
        this.unbindClient();
    }
    attachRequest(response, args, _request) {
        console.log("[attachRequest] called", args);
        this.sendOutput(`establishing the debug session, host: ${args.host} port: ${args.port}`);
        this.cmdQueue = new AsyncBlockingQueue_1.AsyncBlockingQueue();
        this.baseDir = args.basedir;
        this.breakpointMap = {};
        this.server = net.createServer()
            .listen(args.port, args.host)
            .on('listening', () => {
            this.sendOutput(`mobdebug server established, waiting for the lua client to be started`);
        })
            .on('error', e => {
            const errText = `fail to establish the debug server: ${e}`;
            this.sendErrorOutput(errText);
            response.success = false;
            response.message = errText;
            this.sendResponse(response);
        })
            .on('connection', client => {
            this.sendResponse(response);
            // dont set the encoding of the client
            this.onConnected(client);
        });
    }
    disconnectRequest(response, args, request) {
        console.log("[disconnectRequest] called");
        this.sendResponse(response);
        this.onDetach();
    }
    async singleStep(action, threadId) {
        this.client?.write(`${action}\n`);
        const ackLine = await this.clientReader?.readLine();
        this.sendOutput(`[${action}] Ack: ${ackLine}`);
        this.sendEvent(new debugadapter_2.ContinuedEvent(threadId));
        const completeLine = await this.clientReader?.readLine();
        this.sendOutput(`[${action}] Complete: ${completeLine}`);
        this.sendEvent(new debugadapter_2.StoppedEvent('breakpoint', threadId));
    }
    runSingleStepAction(action, response, args, _request) {
        const threadId = args.threadId;
        console.log(`[runSingleStepAction] called, action: ${action} threadId: ${threadId}`);
        this.sendResponse(response);
        this.runInIOLoop(async () => {
            await this.singleStep(action, threadId);
        });
    }
    continueRequest(response, args, request) {
        this.runSingleStepAction("RUN", response, args, request);
    }
    nextRequest(response, args, request) {
        this.runSingleStepAction("OVER", response, args, request);
    }
    stepInRequest(response, args, request) {
        this.runSingleStepAction("STEP", response, args, request);
    }
    stepOutRequest(response, args, request) {
        this.runSingleStepAction("OUT", response, args, request);
    }
    async setBreakPointsRequest(response, args, request) {
        console.log(`[setBreakPointsRequest] called:`, args);
        if (args.source.path === undefined) {
            response.success = false;
            response.message = 'breakpoints has no file path';
            this.sendResponse(response);
            return;
        }
        const path = (0, path_1.normalize)(args.source.path);
        const debuggerPath = path_1.sep === '\\' ? path.replaceAll('\\', '/') : path;
        this.runInIOLoop(async () => {
            // remove old breakpoints of the path
            const bps = this.breakpointMap[debuggerPath] ?? [];
            this.breakpointMap[debuggerPath] = bps;
            for (const line of bps.splice(0)) {
                this.client?.write(`DELB ${debuggerPath} ${line}\n`);
                const delLine = await this.clientReader?.readLine();
                console.log(`[DELB] ${delLine}`);
            }
            const editorBreakPoints = args.breakpoints ?? [];
            const finalBreakpoints = [];
            // set & validate breakpoints of this file
            // todo: check setb error code to validate
            for (const bp of editorBreakPoints) {
                this.client?.write(`SETB ${debuggerPath} ${bp.line}\n`);
                const setLine = await this.clientReader?.readLine();
                console.log(`[SETB] ${setLine}`);
                bps.push(bp.line);
                const nbp = new debugadapter_3.Breakpoint(true, bp.line);
                nbp.setId(this.breakPointCount++);
                finalBreakpoints.push(nbp);
            }
            // response the valid breakpoints
            response.body = { breakpoints: finalBreakpoints, };
            this.sendResponse(response);
        });
    }
    threadsRequest(response, request) {
        console.log("[threadsRequest] called");
        response.body = {
            threads: [
                new debugadapter_3.Thread(1, "thread 1") // thread id is always 1
            ]
        };
        this.sendResponse(response);
    }
    async callStackTrace() {
        if (this.client === null || this.clientReader === null) {
            throw new Error("[STACKHELP] client is disconnected");
        }
        this.client.write('STACKHELP\n');
        const line = await this.clientReader.readLine();
        console.log(`[STACKHELP] ${line}`);
        const matches = /^(\d+)\s+\w+\s+(.+)\s*$/.exec(line);
        if (matches === null) {
            this.sendErrorOutput(`bad response of STACKHELP command`);
            throw new Error("[STACKHELP] bad response");
        }
        if (matches[1] === "200") {
            const frames = matches[2].split(';').filter(s => s.length > 0);
            const stackInfo = frames.map(frame => {
                const [level, path, line, func] = frame.split(',');
                if (!path.startsWith("=")) {
                    const debugPath = path_1.sep === '\\' ? path.replaceAll('/', '\\') : path;
                    const absPath = (0, path_1.isAbsolute)(debugPath) ? debugPath : undefined;
                    const name = (0, path_1.basename)(debugPath);
                    const realName = name === "tolua" ? "-internal-" : name;
                    const realPath = name === "tolua" ? undefined : absPath;
                    return new debugadapter_3.StackFrame(parseInt(level), func, new debugadapter_3.Source(realName, realPath), parseInt(line));
                }
                else {
                    return new debugadapter_3.StackFrame(parseInt(level), func);
                }
            });
            return stackInfo;
        }
        else {
            throw new Error("[STACKHELP] bad response code");
        }
    }
    isTopInternal(stackInfo, threadId) {
        if (stackInfo.length > 0) {
            const name = stackInfo[0].source?.name;
            if (name === "-internal-") {
                return true;
            }
        }
        return false;
    }
    async topInternalOut(stackInfo, threadId, action) {
        if (this.isTopInternal(stackInfo, threadId)) {
            await this.singleStep(action, threadId);
            return true;
        }
        return false;
    }
    async stackTraceRequest(response, { threadId }, request) {
        console.log("[stackTraceRequest] called");
        this.runInIOLoop(async () => {
            let lim = 10;
            while (lim > 0) {
                const stackInfo = await this.callStackTrace();
                lim--;
                let c = await this.topInternalOut(stackInfo, threadId, "OUT");
                if (!c) {
                    break;
                }
            }
            const stackInfo = await this.callStackTrace();
            response.body = { stackFrames: stackInfo, };
            this.sendResponse(response);
        });
    }
    scopesRequest(response, args, request) {
        console.log("[scopesRequest] called");
        // todo
        this.sendResponse(response);
    }
    variablesRequest(response, args, request) {
        console.log("[variablesRequest] called");
        // todo
        this.sendResponse(response);
    }
    evaluateRequest(response, args, request) {
        console.log("[evaluateRequest] called", args);
        this.runInIOLoop(async () => {
            if (this.client === null || this.clientReader === null) {
                return;
            }
            const stackNumber = args.frameId ?? 1;
            this.client.write(`EXEC return ${args.expression.replaceAll('\n', '\r')} -- {stack=${stackNumber}}\n`);
            const line = await this.clientReader.readLine();
            const matches = /^\d+.+?\s+(\d+)\s*$/.exec(line);
            if (matches === null) {
                throw new Error(`[EXEC] bad response: ${line}`);
            }
            const n = parseInt(matches[1]);
            const text = await this.clientReader.readBytesAsString(n);
            response.body = {
                result: text,
                variablesReference: 0,
            };
            // this.sendOutput(text)
            this.sendResponse(response);
        });
    }
    customRequest(command, response, args, request) {
        console.log("[customRequest] called", request);
    }
    async startIOLoop() {
        try {
            while (true) {
                const p = await this.cmdQueue.dequeue();
                try {
                    await p();
                }
                catch (err) {
                    console.error('[startDebuggerLoop] execution error', err, err?.stack);
                }
            }
        }
        catch (reason) {
            console.log('[startDebuggerLoop] caught: ', reason);
        }
        finally {
            console.log('[startDebuggerLoop] stopped');
        }
    }
    stopIOLoop() {
        this.cmdQueue.releaseBlockedEnqueue();
        this.cmdQueue.releaseBlockedDequeue('debugger loop stopped');
    }
    async runInIOLoop(p) {
        await this.cmdQueue.enqueue(p);
    }
    async runInitCommands() {
        this.runInIOLoop(async () => {
            this.sendEvent(new debugadapter_2.InitializedEvent);
            // this.client?.write(`BASEDIR ${this.baseDir}\n`)
            // console.log('[BASEDIR]', await this.clientReader?.readLine())
            console.log('basedir:', this.baseDir);
            this.sendEvent(new debugadapter_2.StoppedEvent('breakpoint', 1));
        });
    }
    async runLastCommands() {
        this.sendEvent(new debugadapter_2.TerminatedEvent);
    }
    sendOutput(text) {
        this.sendEvent(new debugadapter_2.OutputEvent(text + '\n'));
    }
    sendErrorOutput(text) {
        this.sendEvent(new debugadapter_2.OutputEvent(text + '\n', 'stderr'));
    }
}
exports.MobdebugSession = MobdebugSession;
//# sourceMappingURL=MobdebugSession.js.map